@@ -37,6 +37,10 @@
 #include <unistd.h>
 #endif /* HAVE_UNISTD_H */
 
+#if HAVE_SYS_SOCKET_H
+#include <sys/socket.h>
+#endif /* HAVE_SYS_SOCKET_H */
+
 #ifdef _WIN32
 #include <process.h>
 #endif
@@ -48,7 +52,7 @@
 #include <freetds/tls.h>
 #include <freetds/stream.h>
 #include <freetds/checks.h>
-#include "replacements.h"
+#include <freetds/replacements.h>
 
 static TDSRET tds_send_login(TDSSOCKET * tds, const TDSLOGIN * login);
 static TDSRET tds71_do_login(TDSSOCKET * tds, TDSLOGIN * login);
@@ -300,11 +304,53 @@ free_save_context(TDSSAVECONTEXT *ctx)
 }
 
 /**
- * Retrieve and set @@spid
+ * Set @@spid based on column data
+ * \tds
+ * @param curcol  column with spid data.
+ */
+static TDSRET
+tds_set_spid(TDSSOCKET * tds, TDSCOLUMN *curcol)
+{
+	switch (tds_get_conversion_type(curcol->column_type, curcol->column_size)) {
+	case SYBINT2:
+		tds->conn->spid = *((TDS_USMALLINT *) curcol->column_data);
+		break;
+	case SYBINT4:
+		tds->conn->spid = *((TDS_UINT *) curcol->column_data);
+		break;
+	default:
+		return TDS_FAIL;
+	}
+	return TDS_SUCCESS;
+}
+
+/**
+ * Set ncharsize and unicharsize based on column data.
+ * \tds
+ * @param res_info  resultset to get data from.
+ */
+static TDSRET
+tds_set_nvc(TDSSOCKET * tds, TDSRESULTINFO *res_info)
+{
+	int charsize;
+
+	/* Compute the ratios, put some acceptance in order to avoid issues. */
+	/* The "3" constant came from the query issued (NVARCHAR(3) and UNIVARCHAR(3)) */
+	charsize = res_info->columns[0]->on_server.column_size / 3;
+	if (charsize >= 1 && charsize <= 4)
+		tds->conn->ncharsize = (uint8_t) charsize;
+	charsize = res_info->columns[1]->on_server.column_size / 3;
+	if (charsize >= 1 && charsize <= 4)
+		tds->conn->unicharsize = (uint8_t) charsize;
+	return TDS_SUCCESS;
+}
+
+/**
+ * Parse the results from login queries
  * \tds
  */
 static TDSRET
-tds_set_spid(TDSSOCKET * tds)
+tds_parse_login_results(TDSSOCKET * tds)
 {
 	TDS_INT result_type;
 	TDS_INT done_flags;
@@ -317,24 +363,20 @@ tds_set_spid(TDSSOCKET * tds)
 
 		switch (result_type) {
 		case TDS_ROW_RESULT:
-			if (!tds->res_info)
+			if (!tds->res_info && tds->res_info->num_cols < 1)
 				return TDS_FAIL;
-			if (tds->res_info->num_cols != 1)
-				break;
 			curcol = tds->res_info->columns[0];
-			switch (tds_get_conversion_type(curcol->column_type, curcol->column_size)) {
-			case SYBINT2:
-				tds->conn->spid = *((TDS_USMALLINT *) curcol->column_data);
-				break;
-			case SYBINT4:
-				tds->conn->spid = *((TDS_UINT *) curcol->column_data);
-				break;
-			default:
-				return TDS_FAIL;
-			}
+			if (tds->res_info->num_cols == 1 && strcmp(tds_dstr_cstr(&curcol->column_name), "spid") == 0)
+				rc = tds_set_spid(tds, curcol);
+			if (tds->res_info->num_cols == 2 && strcmp(tds_dstr_cstr(&curcol->column_name), "nvc") == 0)
+				rc = tds_set_nvc(tds, tds->res_info);
+			if (TDS_FAILED(rc))
+				return rc;
 			break;
 
 		case TDS_DONE_RESULT:
+		case TDS_DONEPROC_RESULT:
+		case TDS_DONEINPROC_RESULT:
 			if ((done_flags & TDS_DONE_ERROR) != 0)
 				return TDS_FAIL;
 			break;
@@ -352,26 +394,32 @@ tds_setup_connection(TDSSOCKET *tds, TDSLOGIN *login, bool set_db, bool set_spid
 	TDSRET erc;
 	char *str;
 	int len;
+	bool parse_results = false;
 
-	len = 128 + tds_quote_id(tds, NULL, tds_dstr_cstr(&login->database),-1);
+	len = 192 + tds_quote_id(tds, NULL, tds_dstr_cstr(&login->database),-1);
 	if ((str = tds_new(char, len)) == NULL)
 		return TDS_FAIL;
 
 	str[0] = 0;
 	if (login->text_size) {
-		sprintf(str, "set textsize %d ", login->text_size);
+		sprintf(str, "SET TEXTSIZE %d ", login->text_size);
 	}
 	if (set_spid && tds->conn->spid == -1) {
-		strcat(str, "select @@spid ");
+		strcat(str, "SELECT @@spid AS spid ");
+		parse_results = true;
 	}
 	/* Select proper database if specified.
 	 * SQL Anywhere does not support multiple databases and USE statement
 	 * so don't send the request to avoid connection failures */
 	if (set_db && !tds_dstr_isempty(&login->database) &&
 	    (tds->conn->product_name == NULL || strcasecmp(tds->conn->product_name, "SQL Anywhere") != 0)) {
-		strcat(str, "use ");
+		strcat(str, "USE ");
 		tds_quote_id(tds, strchr(str, 0), tds_dstr_cstr(&login->database), -1);
 	}
+	if (IS_TDS50(tds->conn)) {
+		strcat(str, " SELECT CAST('abc' AS NVARCHAR(3)) AS nvc, CAST('xyz' AS UNIVARCHAR(3)) AS uvc");
+		parse_results = true;
+	}
 
 	/* nothing to set, just return */
 	if (str[0] == 0) {
@@ -384,8 +432,8 @@ tds_setup_connection(TDSSOCKET *tds, TDSLOGIN *login, bool set_db, bool set_spid
 	if (TDS_FAILED(erc))
 		return erc;
 
-	if (set_spid && tds->conn->spid == -1)
-		erc = tds_set_spid(tds);
+	if (parse_results)
+		erc = tds_parse_login_results(tds);
 	else
 		erc = tds_process_simple_query(tds);
 
@@ -436,14 +484,10 @@ tds_connect(TDSSOCKET * tds, TDSLOGIN * login, int *p_oserr)
 		const TDSCONTEXT *old_ctx = tds_get_ctx(tds);
 		typedef void (*env_chg_func_t) (TDSSOCKET * tds, int type, char *oldval, char *newval);
 		env_chg_func_t old_env_chg = tds->env_chg_func;
-		/* the context of a socket is const; we have to modify it to suppress error messages during multiple tries. */
-		TDSCONTEXT *mod_ctx = (TDSCONTEXT *) tds_get_ctx(tds);
-		err_handler_t err_handler = tds_get_ctx(tds)->err_handler;
 
 		init_save_context(&save_ctx, old_ctx);
 		tds_set_ctx(tds, &save_ctx.ctx);
 		tds->env_chg_func = tds_save_env;
-		mod_ctx->err_handler = NULL;
 
 		for (i = 0; i < TDS_VECTOR_SIZE(versions); ++i) {
 			int orig_size = tds->conn->env.block_size;
@@ -463,7 +507,6 @@ tds_connect(TDSSOCKET * tds, TDSLOGIN * login, int *p_oserr)
 				break;
 		}
 		
-		mod_ctx->err_handler = err_handler;
 		tds->env_chg_func = old_env_chg;
 		tds_set_ctx(tds, old_ctx);
 		replay_save_context(tds, &save_ctx);
@@ -488,25 +531,12 @@ tds_connect(TDSSOCKET * tds, TDSLOGIN * login, int *p_oserr)
 	tds->login = login;
 
 	tds->conn->tds_version = login->tds_version;
-	tds->conn->emul_little_endian = login->emul_little_endian;
-#ifdef WORDS_BIGENDIAN
-	/*
-	 * Enable automatically little endian emulation.
-	 * TDS 7/8 only supports little endian.
-	 * This is done even for 4.2 to ensure that when we connect to a
-	 * MSSQL we use it and avoid to make additional checks for
-	 * broken 7.0 servers.
-	 * Note that 4.2 should not be used anymore for real jobs.
-	 */
-	if (IS_TDS7_PLUS(tds->conn) || IS_TDS42(tds->conn))
-		tds->conn->emul_little_endian = 1;
-#endif
 
 	/* set up iconv if not already initialized*/
 	if (tds->conn->char_convs[client2ucs2]->to.cd == (iconv_t) -1) {
 		if (!tds_dstr_isempty(&login->client_charset)) {
 			if (TDS_FAILED(tds_iconv_open(tds->conn, tds_dstr_cstr(&login->client_charset), login->use_utf16)))
-				return -TDSEMEM;
+				return -TDSEICONVAVAIL;
 		}
 	}
 
@@ -531,9 +561,39 @@ tds_connect(TDSSOCKET * tds, TDSLOGIN * login, int *p_oserr)
 	tds->conn->capabilities = login->capabilities;
 
 reroute:
+	tds_ssl_deinit(tds->conn);
 	erc = TDSEINTF;
 	orig_port = login->port;
 	for (addrs = login->ip_addrs; addrs != NULL; addrs = addrs->ai_next) {
+
+		/*
+		 * By some reasons ftds forms 3 linked tds_addrinfo (addrs
+		 * variable here) for one server address. The structures
+		 * differs in their ai_socktype and ai_protocol field
+		 * values. Typically the combinations are:
+		 * ai_socktype     | ai_protocol
+		 * -----------------------------
+		 * 1 (SOCK_STREAM) | 6  (tcp)
+		 * 2 (SOCK_DGRAM)  | 17 (udp)
+		 * 3 (SOCK_RAW)    | 0  (ip)
+		 *
+		 * Later on these fields are not used and dtds always
+		 * creates a tcp socket. In case if there is a connection
+		 * problem this behavior leads to 3 tries with the provided
+		 * timeout which basically multiplies the spent time
+		 * without any good result. So it was decided to skip the
+		 * non tcp addresses.
+		 *
+		 * NOTE: on Windows exactly one tds_addrinfo structure is
+		 *	 formed and it has 0 in both ai_socktype and
+		 *	 ai_protocol fields. So skipping is conditional for
+		 *	 non-Windows platforms
+		 */
+#ifndef _WIN32
+		if (addrs->ai_socktype != SOCK_STREAM)
+			continue;
+#endif
+
 		login->port = orig_port;
 
 		if (!IS_TDS50(tds->conn) && !tds_dstr_isempty(&login->instance_name) && !login->port)
@@ -624,9 +684,28 @@ reroute:
 #if ENABLE_ODBC_MARS
 	/* initialize SID */
 	if (IS_TDS72_PLUS(tds->conn) && login->mars) {
-		tds->conn->sessions[0] = NULL;
+		TDS72_SMP_HEADER *p;
+
+		tds_extra_assert(tds->sid == 0);
+		tds_extra_assert(tds->conn->sessions[0] == tds);
+		tds_extra_assert(tds->send_packet != NULL);
+		tds_extra_assert(!tds->send_packet->next);
+
 		tds->conn->mars = 1;
-		tds->sid = -1;
+
+		/* start session with a SMP SYN */
+		if (TDS_FAILED(tds_append_syn(tds)))
+			return -TDSEMEM;
+
+		/* reallocate send_packet */
+		if (!tds_realloc_socket(tds, tds->out_buf_max))
+			return -TDSEMEM;
+
+		/* start SMP DATA header */
+		p = (TDS72_SMP_HEADER *) tds->send_packet->buf;
+		p->signature = TDS72_SMP;
+		p->type = TDS_SMP_DATA;
+
 		tds_init_write_buf(tds);
 	}
 #endif
@@ -660,10 +739,6 @@ tds_put_login_string(TDSSOCKET * tds, const char *buf, int n)
 static TDSRET
 tds_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 {
-#ifdef WORDS_BIGENDIAN
-	static const unsigned char be1[] = { 0x02, 0x00, 0x06, 0x04, 0x08, 0x01 };
-	static const unsigned char be2[] = { 0x00, 12, 16 };
-#endif
 	static const unsigned char le1[] = { 0x03, 0x01, 0x06, 0x0a, 0x09, 0x01 };
 	static const unsigned char le2[] = { 0x00, 13, 17 };
 
@@ -684,6 +759,8 @@ tds_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 
 	unsigned char protocol_version[4];
 	unsigned char program_version[4];
+	unsigned char sec_flags = 0;
+	bool use_kerberos = false;
 
 	int len;
 	char blockstr[16];
@@ -698,13 +775,27 @@ tds_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 		return TDS_FAIL;
 	}
 	if (tds_dstr_isempty(&login->user_name)) {
-		tdsdump_log(TDS_DBG_ERROR, "Kerberos login not supported using TDS 4.x or 5.0\n");
+		if (!IS_TDS50(tds->conn)) {
+			tdsdump_log(TDS_DBG_ERROR, "Kerberos login not supported using TDS 4.x\n");
+			return TDS_FAIL;
+		}
+
+#ifdef ENABLE_KRB5
+		/* try kerberos */
+		sec_flags = TDS5_SEC_LOG_SECSESS;
+		use_kerberos = true;
+		tds->conn->authentication = tds_gss_get_auth(tds);
+		if (!tds->conn->authentication)
+			return TDS_FAIL;
+#else
+		tdsdump_log(TDS_DBG_ERROR, "requested GSS authentication but not compiled in\n");
 		return TDS_FAIL;
+#endif
 	}
 	if (encryption_level == TDS_ENCRYPTION_DEFAULT)
 		encryption_level = TDS_ENCRYPTION_OFF;
-	if (encryption_level != TDS_ENCRYPTION_OFF) {
-		if (IS_TDS42(tds->conn)) {
+	if (!use_kerberos && encryption_level != TDS_ENCRYPTION_OFF) {
+		if (!IS_TDS50(tds->conn)) {
 			tdsdump_log(TDS_DBG_ERROR, "Encryption not supported using TDS 4.x\n");
 			return TDS_FAIL;
 		}
@@ -742,15 +833,7 @@ tds_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	}
 	sprintf(blockstr, "%d", (int) getpid());
 	tds_put_login_string(tds, blockstr, TDS_MAXNAME);	/* host process */
-#ifdef WORDS_BIGENDIAN
-	if (tds->conn->emul_little_endian) {
-		tds_put_n(tds, le1, 6);
-	} else {
-		tds_put_n(tds, be1, 6);
-	}
-#else
 	tds_put_n(tds, le1, 6);
-#endif
 	tds_put_byte(tds, !login->bulk_copy);
 	tds_put_n(tds, NULL, 2);
 	if (IS_TDS42(tds->conn)) {
@@ -783,22 +866,16 @@ tds_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	} else {
 		tds_put_n(tds, program_version, 4);	/* program version ? */
 	}
-#ifdef WORDS_BIGENDIAN
-	if (tds->conn->emul_little_endian) {
-		tds_put_n(tds, le2, 3);
-	} else {
-		tds_put_n(tds, be2, 3);
-	}
-#else
 	tds_put_n(tds, le2, 3);
-#endif
 	tds_put_login_string(tds, tds_dstr_cstr(&login->language), TDS_MAXNAME);	/* language */
 	tds_put_byte(tds, login->suppress_language);
 
 	/* oldsecure(2), should be zero, used by old software */
 	tds_put_n(tds, NULL, 2);
 	/* seclogin(1) bitmask */
-	tds_put_byte(tds, encryption_level != TDS_ENCRYPTION_OFF ? TDS5_SEC_LOG_ENCRYPT2|TDS5_SEC_LOG_NONCE : 0);
+	if (sec_flags == 0 && encryption_level != TDS_ENCRYPTION_OFF)
+		sec_flags = TDS5_SEC_LOG_ENCRYPT2|TDS5_SEC_LOG_ENCRYPT3;
+	tds_put_byte(tds, sec_flags);
 	/* secbulk(1)
 	 * halogin(1) type of ha login
 	 * hasessionid(6) id of session to reconnect
@@ -826,11 +903,18 @@ tds_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	} else if (IS_TDS50(tds->conn)) {
 		/* just padding to 8 bytes */
 		tds_put_n(tds, NULL, 4);
+
+		/* send capabilities */
 		tds_put_byte(tds, TDS_CAPABILITY_TOKEN);
 		tds_put_smallint(tds, sizeof(tds->conn->capabilities));
 		tds_put_n(tds, &tds->conn->capabilities, sizeof(tds->conn->capabilities));
 	}
 
+#ifdef ENABLE_KRB5
+	if (use_kerberos)
+		tds5_gss_send(tds);
+#endif
+
 	return tds_flush_packet(tds);
 }
 
@@ -880,6 +964,11 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	size_t user_name_len = strlen(user_name);
 	size_t auth_len = 0;
 
+	static const char ext_data[] =
+		"\x0a\x01\x00\x00\x00\x01"	/* Enable UTF-8 */
+		"\xff";
+	size_t ext_len = IS_TDS74_PLUS(tds->conn) ? sizeof(ext_data) - 1 : 0;
+
 	/* fields */
 	enum {
 		HOST_NAME,
@@ -887,6 +976,7 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 		PASSWORD,
 		APP_NAME,
 		SERVER_NAME,
+		EXTENSION,
 		LIBRARY_NAME,
 		LANGUAGE,
 		DATABASE_NAME,
@@ -965,6 +1055,8 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 		option_flag3 |= TDS_CHANGE_PASSWORD;
 		SET_FIELD_DSTR(NEW_PASSWORD, login->new_password, 128);
 	}
+	if (ext_len)
+		option_flag3 |= TDS_EXTENSION;
 
 	/* convert data fields */
 	for (field = data_fields; field < data_fields + TDS_VECTOR_SIZE(data_fields); ++field) {
@@ -979,6 +1071,13 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 				free(data);
 				return TDS_FAIL;
 			}
+		} else if (ext_len && field == &data_fields[EXTENSION]) {
+			if (data_stream.stream.write(&data_stream.stream, 4) != 4) {
+				free(data);
+				return TDS_FAIL;
+			}
+			field->len = 4;
+			continue;
 		}
 		data_stream.size = MIN(data_stream.size, data_pos + field->limit);
 		data_stream.stream.write(&data_stream.stream, 0);
@@ -989,10 +1088,16 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	pwd = (unsigned char *) data + data_fields[NEW_PASSWORD].pos - current_pos;
 	tds7_crypt_pass(pwd, data_fields[NEW_PASSWORD].len, pwd);
 	packet_size += data_stream.size;
+	if (ext_len) {
+		packet_size += ext_len;
+		pwd = (unsigned char *) data + data_fields[EXTENSION].pos - current_pos;
+		TDS_PUT_UA4LE(pwd, current_pos + data_stream.size + auth_len);
+	}
 
 #if !defined(TDS_DEBUG_LOGIN)
 	tdsdump_log(TDS_DBG_INFO2, "quietly sending TDS 7+ login packet\n");
-	tdsdump_off();
+	do { TDSDUMP_OFF_ITEM off_item;
+	tdsdump_off(&off_item);
 #endif
 	TDS_PUT_INT(tds, packet_size);
 	switch (login->tds_version) {
@@ -1060,10 +1165,8 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	/* host name */
 	PUT_STRING_FIELD_PTR(HOST_NAME);
 	if (tds->conn->authentication) {
-		tds_put_smallint(tds, 0);
-		tds_put_smallint(tds, 0);
-		tds_put_smallint(tds, 0);
-		tds_put_smallint(tds, 0);
+		tds_put_int(tds, 0);
+		tds_put_int(tds, 0);
 	} else {
 		/* username */
 		PUT_STRING_FIELD_PTR(USER_NAME);
@@ -1074,9 +1177,13 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	PUT_STRING_FIELD_PTR(APP_NAME);
 	/* server name */
 	PUT_STRING_FIELD_PTR(SERVER_NAME);
-	/* unknown */
-	tds_put_smallint(tds, 0);
-	tds_put_smallint(tds, 0);
+	/* extensions */
+	if (ext_len) {
+		TDS_PUT_SMALLINT(tds, data_fields[EXTENSION].pos);
+		tds_put_smallint(tds, 4);
+	} else {
+		tds_put_int(tds, 0);
+	}
 	/* library name */
 	PUT_STRING_FIELD_PTR(LIBRARY_NAME);
 	/* language  - kostya@warmcat.excom.spb.su */
@@ -1090,7 +1197,7 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 
 	/* authentication stuff */
 	TDS_PUT_SMALLINT(tds, current_pos + data_stream.size);
-	TDS_PUT_SMALLINT(tds, auth_len);	/* this matches numbers at end of packet */
+	TDS_PUT_SMALLINT(tds, MIN(auth_len, 0xffffu));
 
 	/* db file */
 	PUT_STRING_FIELD_PTR(DB_FILENAME);
@@ -1100,7 +1207,7 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 		PUT_STRING_FIELD_PTR(NEW_PASSWORD);
 
 		/* SSPI long */
-		tds_put_int(tds, 0);
+		tds_put_int(tds, auth_len >= 0xffffu ? auth_len : 0);
 	}
 
 	tds_put_n(tds, data, data_stream.size);
@@ -1108,8 +1215,15 @@ tds7_send_login(TDSSOCKET * tds, const TDSLOGIN * login)
 	if (tds->conn->authentication)
 		tds_put_n(tds, tds->conn->authentication->packet, auth_len);
 
+	if (ext_len)
+		tds_put_n(tds, ext_data, ext_len);
+
 	rc = tds_flush_packet(tds);
-	tdsdump_on();
+
+#if !defined(TDS_DEBUG_LOGIN)
+	tdsdump_on(&off_item);
+	} while(0);
+#endif
 
 	free(data);
 	return rc;
@@ -1166,6 +1280,7 @@ tds71_do_login(TDSSOCKET * tds, TDSLOGIN* login)
 	SET_UI16BE(13, instance_name_len);
 	if (!IS_TDS72_PLUS(tds->conn)) {
 		SET_UI16BE(16, START_POS + 6 + 1 + instance_name_len);
+		/* strip MARS setting */
 		buf[20] = 0xff;
 	} else {
 		start_pos += 5;
@@ -1269,7 +1384,7 @@ tds71_do_login(TDSSOCKET * tds, TDSLOGIN* login)
 	tds->in_pos += pkt_len;
 	/* TODO some mssql version do not set last packet, update tds according */
 
-	tdsdump_log(TDS_DBG_INFO1, "detected flag %d\n", crypt_flag);
+	tdsdump_log(TDS_DBG_INFO1, "detected crypt flag %d\n", crypt_flag);
 
 	/* if server do not has certificate do normal login */
 	if (crypt_flag == TDS7_ENCRYPT_NOT_SUP) {
@@ -1281,7 +1396,7 @@ tds71_do_login(TDSSOCKET * tds, TDSLOGIN* login)
 	}
 
 	/*
-	 * if server has a certificate it require at least a crypted login
+	 * if server has a certificate it requires at least a crypted login
 	 * (even if data is not encrypted)
 	 */
 
@@ -1297,7 +1412,7 @@ tds71_do_login(TDSSOCKET * tds, TDSLOGIN* login)
 
 	ret = tds7_send_login(tds, login);
 
-	/* if flag is 0 it means that after login server continue not encrypted */
+	/* if flag is TDS7_ENCRYPT_OFF(0) it means that after login server continue not encrypted */
 	if (crypt_flag == TDS7_ENCRYPT_OFF || TDS_FAILED(ret))
 		tds_ssl_deinit(tds->conn);
 
